package org.xbib.event.async.impl;

import org.xbib.event.async.Async;
import org.xbib.event.async.AsyncOptions;
import org.xbib.event.async.spi.AsyncThreadFactory;
import org.xbib.event.async.spi.ExecutorServiceFactory;
import org.xbib.event.loop.EventLoopGroup;

/**
 * Async builder for creating instances with SPI overrides.
 */
public class AsyncBuilder {

  private final AsyncOptions options;
  private AsyncThreadFactory threadFactory;
  private ExecutorServiceFactory executorServiceFactory;

  private EventLoopGroup eventLoopGroup;

  public AsyncBuilder() {
    this(new AsyncOptions());
  }

  public AsyncBuilder(AsyncOptions options) {
    this.options = options;
  }

  /**
   * @return the options
   */
  public AsyncOptions options() {
    return options;
  }

  /**
   * @return the {@code AsyncThreadFactory} to use
   */
  public AsyncThreadFactory threadFactory() {
    return threadFactory;
  }

  /**
   * Set the {@code AsyncThreadFactory} instance to use.
   * @param factory the metrics
   * @return this builder instance
   */
  public AsyncBuilder threadFactory(AsyncThreadFactory factory) {
    this.threadFactory = factory;
    return this;
  }

  /**
   * @return the {@code ExecutorServiceFactory} to use
   */
  public ExecutorServiceFactory executorServiceFactory() {
    return executorServiceFactory;
  }

  /**
   * Set the {@code ExecutorServiceFactory} instance to use.
   * @param factory the factory
   * @return this builder instance
   */
  public AsyncBuilder executorServiceFactory(ExecutorServiceFactory factory) {
    this.executorServiceFactory = factory;
    return this;
  }

  public EventLoopGroup eventLoopGroup() {
    return eventLoopGroup;
  }

  public AsyncBuilder eventLoopGroup(EventLoopGroup eventLoopGroup) {
    this.eventLoopGroup = eventLoopGroup;
    return this;
  }

  /**
   * Build and return the instance
   */
  public Async newInstance() {
    AsyncImpl async = new AsyncImpl(options, threadFactory, executorServiceFactory, eventLoopGroup);
    async.init();
    return async;
  }

  /**
   * Initialize the service providers.
   * @return this builder instance
   */
  public AsyncBuilder init() {
    initThreadFactory();
    initExecutorServiceFactory();
    return this;
  }

  private void initThreadFactory() {
    if (threadFactory != null) {
      return;
    }
    threadFactory = AsyncThreadFactory.INSTANCE;
  }

  private void initExecutorServiceFactory() {
    if (executorServiceFactory != null) {
      return;
    }
    executorServiceFactory = ExecutorServiceFactory.INSTANCE;
  }
}
