package org.xbib.event.async.io;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.AsynchronousFileChannel;
import java.nio.channels.CompletionHandler;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.security.InvalidParameterException;
import java.util.function.ObjIntConsumer;

import static java.nio.channels.AsynchronousFileChannel.open;

/**
 * Asynchronous non-blocking read operations that use an underlying AsynchronousFileChannel.
 */
public abstract class AbstractAsyncFileReaderLines {

    static final int BUFFER_SIZE = 8192;

    private static final int MAX_LINE_SIZE = 4096;

    private static final int LF = '\n';

    private static final int CR = '\r';

    private boolean cancelled = false;

    public AbstractAsyncFileReaderLines() {
    }

    protected abstract void onError(Throwable error);

    protected abstract void onComplete();

    protected abstract void onProduceLine(String line);

    public final void cancel() {
        cancelled = true;
    }

    public final boolean isCancelled() {
        return cancelled;
    }

    final void readLines(Path file) throws IOException {
        readLines(open(file, StandardOpenOption.READ), BUFFER_SIZE);
    }

    /**
     * Read bytes from an {@code AsynchronousFileChannel}, which are decoded into characters
     * using the UTF-8 charset.
     * The resulting characters are parsed by line and passed to the destination buffer.
     *
     * @param asyncFile  the nio associated file channel.
     * @param bufferSize buffer size
     */
    final void readLines(AsynchronousFileChannel asyncFile, int bufferSize) {
        readLines(asyncFile, 0, 0, 0, new byte[bufferSize], new byte[MAX_LINE_SIZE], 0);
    }

    /**
     * There is a recursion on `readLines()`establishing a serial order among:
     * `readLines()` -> `produceLine()` -> `onProduceLine()` -> `readLines()` -> and so on.
     * It finishes with a call to `close()`.
     *
     * @param asyncFile the nio associated file channel.
     * @param position  current read or write position in file.
     * @param bufpos    read position in buffer.
     * @param bufsize   total bytes in buffer.
     * @param buffer    buffer for current producing line.
     * @param auxline   the transfer buffer.
     * @param linepos   current position in producing line.
     */
    private void readLines(AsynchronousFileChannel asyncFile, long position, int bufpos,
            int bufsize, byte[] buffer, byte[] auxline, int linepos)
    {
        while (bufpos < bufsize) {
            if (buffer[bufpos] == LF) {
                if (linepos > 0 && auxline[linepos - 1] == CR) linepos--;
                bufpos++;
                produceLine(auxline, linepos);
                linepos = 0;
            } else if (linepos == MAX_LINE_SIZE - 1) {
                produceLine(auxline, linepos);
                linepos = 0;
            } else auxline[linepos++] = buffer[bufpos++];
        }
        int lastLinePos = linepos;
        if (!isCancelled()) readBytes(asyncFile, position, buffer, buffer.length, (err, res) -> {
            if (isCancelled())
                return;
            if (err != null) {
                onError(err);
                close(asyncFile);
                return;
            }
            if (res <= 0) {
                if (lastLinePos > 0) {
                    produceLine(auxline, lastLinePos);
                }
                close(asyncFile);
            } else {
                readLines(asyncFile, position + res, 0, res, buffer, auxline, lastLinePos);
            }
        });
    }

    private void close(AsynchronousFileChannel asyncFile) {
        try {
            asyncFile.close();
        } catch (IOException e) {
            onError(e);
        } finally {
            onComplete();
        }
    }

    /**
     * This is called only from readLines() callback and performed from a background IO thread.
     *
     * @param auxline the transfer buffer.
     * @param linepos current position in producing line.
     */
    private void produceLine(byte[] auxline, int linepos) {
        String line = new String(auxline, 0, linepos, StandardCharsets.UTF_8);
        onProduceLine(line);
    }

    /**
     * Asynchronous read chunk operation, callback based.
     */
    private static void readBytes(AsynchronousFileChannel asyncFile,
            long position, byte[] data, int size, ObjIntConsumer<Throwable> completed) {
        if (completed == null) {
            throw new InvalidParameterException("callback can't be null!");
        }
        if (size > data.length) {
            size = data.length;
        }
        if (size == 0) {
            completed.accept(null, 0);
            return;
        }
        ByteBuffer buf = ByteBuffer.wrap(data, 0, size);
        CompletionHandler<Integer, Object> readCompleted = new CompletionHandler<>() {
                    @Override
                    public void completed(Integer result, Object attachment) {
                        completed.accept(null, result);
                    }

                    @Override
                    public void failed(Throwable exc, Object attachment) {
                        completed.accept(exc, 0);
                    }
                };
        asyncFile.read(buf, position, null, readCompleted);
    }
}
