package org.xbib.event.async.io;

import io.reactivex.rxjava3.core.Observable;
import org.junit.jupiter.api.Test;
import org.reactivestreams.Publisher;
import org.reactivestreams.Subscription;
import org.xbib.event.async.yield.AsyncQuery;

import java.io.FileWriter;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Pattern;
import java.util.stream.Stream;

import static io.reactivex.rxjava3.core.Observable.fromArray;
import static java.lang.System.lineSeparator;
import static java.lang.System.out;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.nio.file.Files.delete;
import static java.util.Arrays.asList;
import static java.util.Collections.max;
import static java.util.Comparator.comparingInt;
import static java.util.stream.Collectors.joining;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;
import static org.xbib.event.async.io.Subscribers.doOnNext;

public class AsyncFileReaderTest {

    private static final Pattern NEWLINE = Pattern.compile("(\r\n|\n|\r)");

    private static final String OUTPUT = "build/output.txt";

    private static final Path METAMORPHOSIS = getPath("Metamorphosis-by-Franz-Kafka.txt");

    private static final Path WIZARD = getPath("The-Wizard-by-Rider-Haggard.txt");

    private static Path getPath(String resource) {
        try {
            return Paths.get(AsyncFileReaderTest.class.getResource("/" + resource).toURI());
        } catch (NullPointerException | URISyntaxException e) {
            e.printStackTrace();
            throw new IllegalStateException();
        }
    }

    public void readLinesAndPrint() throws IOException {
        List<String> expected = asList("super", "brave", "isel", "ole", "gain", "massi", "tot");
        Path path = Paths.get(OUTPUT);
        Files.write(path, expected);
        CompletableFuture<?> cf = new CompletableFuture<>();
        AsyncFiles
                .lines(path)
                .subscribe(doOnNext(out::println)
                        .doOnComplete(() -> cf.complete(null))
                        .doOnSubscribe(sign -> sign.request(Long.MAX_VALUE)));
        cf.join();
    }

    public void asynQueryLinesAndPrint() throws IOException {
        List<String> expected = asList("super", "brave", "isel", "ole", "gain", "massi", "tot");
        Path path = Paths.get(OUTPUT);
        Files.write(path, expected);
        AsyncFiles
                .asyncQuery(path)
                .subscribe((line, err) -> out.println(line))
                .join();
    }

    @Test
    public void readLinesWith4BytesBuffer() throws IOException {
        final List<String> expected = asList("super", "brave", "isel", "ole", "gain", "massi", "tot");
        try (FileWriter out = new FileWriter(OUTPUT)) {
            expected.forEach(item -> writeLine(out, item));
            out.flush();
        }
        try {
            CompletableFuture<Void> p = new CompletableFuture<>();
            Iterator<String> iter = expected.iterator();
            AsyncFiles
                    .lines(4, OUTPUT)
                    .subscribe(doOnNext(item -> {
                        if (p.isDone()) return;
                        String curr = iter.next();
                        assertEquals(curr, item);
                    })
                            .doOnSubscribe(sign -> sign.request(Integer.MAX_VALUE))
                            .doOnError(p::completeExceptionally)
                            .doOnComplete(() -> p.complete(null)));
            p.join();
            assertFalse(iter.hasNext(), "Missing items not retrieved by lines subscriber!!");
        } finally {
            delete(Paths.get(OUTPUT));
        }
    }

    private void writeLine(FileWriter out, String item) {
        try {
            out.write(item + '\n');
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    @Test
    public void readAllBytesWith8BytesBuffer() throws IOException {
        final List<String> expected = asList("super", "brave", "isel", "ole", "gain", "massi", "tot");
        Files.write(Paths.get(OUTPUT), expected);
        try {
            AsyncFiles
                    .readAll(OUTPUT, 8)
                    .thenApply(NEWLINE::splitAsStream)
                    .thenApply(Stream::iterator)
                    .thenAccept(actual -> expected.forEach(l -> {
                        if (!actual.hasNext()) {
                            fail("File does not contain line: " + l);
                        }
                        String curr = actual.next();
                        assertEquals(l, curr);
                    }))
                    .join();
        } finally {
            delete(Paths.get(OUTPUT));
        }
    }

    @Test
    public void readAllBytesFromLargeFile() throws IOException {
        Path path = METAMORPHOSIS;
        Iterator<String> expected = Files.lines(path, UTF_8).iterator();
        AsyncFiles
                .readAll(path.toString()) // KEEP like this with toString() to force invocation chain cover more methods
                .thenAccept(actual -> NEWLINE
                        .splitAsStream(actual)
                        .forEach(line -> {
                            if (!expected.hasNext())
                                fail("More items read than expected!");
                            assertEquals(expected.next(), line);
                        }))
                .join();
        if (expected.hasNext())
            fail("There are missing lines to read: " + expected.next());

    }

    @Test
    public void readLinesFromLargeFile() throws IOException {
        Path PATH = WIZARD;
        Iterator<String> expected = Files.lines(PATH, UTF_8).iterator();
        CompletableFuture<Void> completed = new CompletableFuture<>();
        AsyncFiles
                .lines(PATH.toString())
                .subscribe(doOnNext(item -> {
                    if (completed.isDone()) return;
                    String curr = expected.next();
                    assertEquals(curr, item);
                })
                        .doOnSubscribe(sign -> sign.request(Integer.MAX_VALUE))
                        .doOnError(completed::completeExceptionally)
                        .doOnComplete(() -> completed.complete(null)));
        completed.join();
        assertFalse(expected.hasNext(), "Missing items not retrieved by lines subscriber!!");
    }

    @Test
    public void readLinesFromLargeFileAndSleep() throws IOException, InterruptedException {
        Path path = WIZARD;
        Iterator<String> expected = Files.lines(path, UTF_8).iterator();
        CompletableFuture<Void> completed = new CompletableFuture<>();
        CompletableFuture<Subscription> sign = new CompletableFuture<>();
        AsyncFiles
                .lines(path.toString())
                .subscribe(doOnNext(item -> {
                    if (completed.isDone()) return;
                    String curr = expected.next();
                    assertEquals(curr, item);
                }).doOnSubscribe(s -> {
                    s.request(400);
                    sign.complete(s);
                }).doOnError(completed::completeExceptionally)
                        .doOnComplete(() -> completed.complete(null)));
        Thread.sleep(1000);
        sign.join().request(Long.MAX_VALUE);
        completed.join();
        assertFalse(expected.hasNext(), "Missing items not retrieved by lines subscriber!!");
    }

    @Test
    public void readLinesFromLargeFileAndCancelation() throws InterruptedException {
        CompletableFuture<Void> completed = new CompletableFuture<>();
        CompletableFuture<Subscription> subscribed = new CompletableFuture<>();
        int[] count = {0};
        AsyncFiles
                .lines(64, WIZARD)
                .subscribe(doOnNext(item -> {
                    count[0]++;
                    if (count[0] > 400) {
                        subscribed.join().cancel();
                        completed.complete(null);
                    }
                })
                        .doOnSubscribe(s -> {
                            s.request(Integer.MAX_VALUE);
                            subscribed.complete(s);
                        })
                        .doOnError(completed::completeExceptionally));
        completed.join();
        Thread.sleep(500); // Wait a little to check that no further signals were emitted in the meanwhile.
        assertEquals(count[0], 401);
    }

    @Test
    public void readLinesFromLargeFileRequestRandomly() throws IOException {
        Path PATH = WIZARD;
        Iterator<String> expected = Files.lines(PATH, UTF_8).iterator();
        Random rand = new Random();
        CompletableFuture<Void> completed = new CompletableFuture<>();
        Subscription[] sign = {null};
        AtomicInteger signals = new AtomicInteger();
        AtomicInteger prevRequest = new AtomicInteger();
        AsyncFiles
                .lines(PATH.toString())
                .subscribe(doOnNext(item -> {
                    if (completed.isDone()) return;
                    String curr = expected.next();
                    assertEquals(curr, item);
                    if (prevRequest.decrementAndGet() == 0) {
                        int req = rand.nextInt(50) + 1;
                        sign[0].request(req);
                        prevRequest.addAndGet(req);
                        signals.incrementAndGet();
                    }
                })
                        .doOnSubscribe(s -> {
                            int req = rand.nextInt(50) + 1;
                            s.request(req);
                            prevRequest.addAndGet(req);
                            sign[0] = s;
                        })
                        .doOnError(completed::completeExceptionally)
                        .doOnComplete(() -> completed.complete(null)));
        completed.join();
        assertFalse(expected.hasNext(), "Missing items not retrieved by lines subscriber!!");
        assertTrue(signals.get() < 2000);
    }

    @Test
    public void readAllBytesFromLargeFileViaCallbackConcurrently() throws IOException {
        Path path = METAMORPHOSIS;
        String expected = Files
                .lines(path, UTF_8)
                .map(line -> line + lineSeparator())
                .collect(joining());
        CompletableFuture<Void> p1 = new CompletableFuture<>();
        CompletableFuture<Void> p2 = new CompletableFuture<>();
        AsyncFiles
                .readAll(path.toString(), (err, actual) -> {
                    assertEquals(expected, actual);
                    p1.complete(null);
                });
        AsyncFiles
                .readAll(path.toString(), (err, actual) -> {
                    assertEquals(expected, actual);
                    p2.complete(null);
                });
        p1.join();
        p2.join();
    }

    @Test
    public void readLinesFromLargeFileRxJavaObservable() throws IOException {
        Path PATH = METAMORPHOSIS;
        Iterator<String> expected = Files.lines(PATH, UTF_8).iterator();
        Observable
                .fromPublisher(AsyncFiles.lines(PATH))
                .doOnError(ex -> fail(ex.getMessage()))
                .doOnNext(line -> {
                    if (!expected.hasNext()) {
                        fail("More items read than expected!");
                    }
                    String next = expected.next();
                    assertEquals(next, line);
                })
                .blockingSubscribe();
        if (expected.hasNext())
            fail("There are missing lines to read: " + expected.next());
    }

    @Test
    public void readLinesFromLargeFileRxJavaObservableAndCountWords() {
        final int MIN = 5;
        final int MAX = 10;
        ConcurrentHashMap<String, Integer> words = new ConcurrentHashMap<>();
        Publisher<String> lines = AsyncFiles.lines(WIZARD);
        Observable
                .fromPublisher(lines)
                .filter(line -> !line.isEmpty())
                .skip(14)
                .takeWhile(line -> !line.contains("*** END OF "))
                .flatMap(line -> fromArray(line.split(" ")))
                .filter(word -> word.length() > MIN && word.length() < MAX)
                .doOnNext(w -> words.merge(w, 1, Integer::sum))
                .blockingSubscribe();
        Map.Entry<String, ? extends Number> common = max(words.entrySet(), comparingInt(Map.Entry::getValue));
        assertEquals(common.getKey(), "Hokosa");
        assertEquals(common.getValue().intValue(), 183);
    }

    /**
     * Trying to reproduce an erroneous behavior where takeWhile tries
     * to terminate upstream subscription before it exist.
     */
    @Test
    public void readLinesAndImmediatelyInterrupt() {
        AsyncFiles
                .asyncQuery(WIZARD)
                .filter(line -> !line.isEmpty())
                .takeWhile(line -> false) // Will finish on first line
                .blockingSubscribe();
    }

    @Test
    public void readLinesFromLargeFileAsyncQueryAndCountWords() {
        final int MIN = 5;
        final int MAX = 10;
        ConcurrentHashMap<String, Integer> words = new ConcurrentHashMap<>();
        AsyncFiles
                .asyncQuery(WIZARD)
                .filter(line -> !line.isEmpty())
                .skip(14)
                .takeWhile(line -> !line.contains("*** END OF "))
                .flatMapMerge(line -> AsyncQuery.of(line.split(" ")))
                .filter(word -> word.length() > MIN && word.length() < MAX)
                .onNext((w, err) -> words.merge(w, 1, Integer::sum))
                .blockingSubscribe();
        Map.Entry<String, ? extends Number> common = max(words.entrySet(), comparingInt(Map.Entry::getValue));
        assertEquals(common.getKey(), "Hokosa");
        assertEquals(common.getValue().intValue(), 183);
    }

    @Test
    public void readLinesFromLargeFileAsyncQueryAndCountDistinctWords() {
        int[] count = {0};
        AsyncFiles
                .asyncQuery(WIZARD)
                .filter(line -> !line.isEmpty())
                .skip(14)
                .takeWhile(line -> !line.contains("*** END OF "))
                .flatMapMerge(line -> AsyncQuery.of(line.split("\\W+")))
                .distinct()
                .onNext((word, err) -> {
                    if (err == null) count[0]++;
                })
                .onNext((word, err) -> {
                    if (err != null) {
                        err.printStackTrace();
                    }
                })
                .blockingSubscribe();
        assertEquals(count[0], 5206);
    }


    /**
     * Despite this test being very similar to the previous one it suppresses the takeWhile()
     * operation from the pipeline.
     * The takeWhile() was hiding a bug in asyncQuery() on completion of the CF.
     */
    @Test
    public void readLinesFromLargeFileAsyncQueryWithoutTakeWhile() {
        final int MIN = 5;
        final int MAX = 10;
        ConcurrentHashMap<String, Integer> words = new ConcurrentHashMap<>();
        AsyncFiles
                .asyncQuery((WIZARD))
                .filter(line -> !line.isEmpty())
                .skip(14)
                .flatMapMerge(line -> AsyncQuery.of(line.split(" ")))
                .filter(word -> word.length() > MIN && word.length() < MAX)
                .onNext((w, err) -> words.merge(w, 1, Integer::sum))
                .blockingSubscribe();
        Map.Entry<String, ? extends Number> common = max(words.entrySet(), comparingInt(Map.Entry::getValue));
        assertEquals(common.getKey(), "Hokosa");
        assertEquals(common.getValue().intValue(), 183);
    }
}