package org.xbib.event.async.io;

import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import static java.nio.file.Files.delete;
import static java.nio.file.Files.lines;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.fail;

public class AsyncFileWriterTest {

    @Test
    public void writeLinesTest() throws IOException {
        final String path = "build/output1.txt";
        final List<String> expected = Arrays.asList("super", "brave", "isel", "ole", "gain", "massive");
        try {
            AsyncFiles.write(Paths.get(path), expected)
                    .whenComplete((index, ex) -> {
                        if (ex != null) {
                            fail(ex.getMessage());
                        }
                    })
                    .join();
            Iterator<String> actual = lines(Paths.get(path)).iterator();
            if (!actual.hasNext()) {
                fail("File is empty!!!");
            }
            expected.forEach(l -> {
                if (actual.hasNext()) {
                    assertEquals(l, actual.next());
                }
                else {
                    fail("File does not contain line: " + l);
                }
            });
        } finally {
            delete(Paths.get(path));
        }
    }

    @Test
    public void writeBytesTest() throws IOException, URISyntaxException {
        final String output = "build/output2.txt";
        URL resource = getClass().getResource("/Metamorphosis-by-Franz-Kafka.txt");
        Path path = Paths.get(resource.toURI());
        byte[] expected = Files.readAllBytes(path);
        AsyncFiles.writeBytes(Paths.get(output), expected)
                .join();
        try {
            AsyncFiles
                    .readAllBytes(Paths.get(output))
                    .whenComplete((actual, ex) -> {
                        if (ex != null) {
                            fail(ex.getMessage());
                        }
                        assertArrayEquals(expected, actual);
                    })
                    .join();
        } finally {
            delete(Paths.get(output));
        }
    }

    @Test
    public void readWriteTestAsyncForReadme() throws IOException, URISyntaxException {
        URL resource = getClass().getResource("/Metamorphosis-by-Franz-Kafka.txt");
        Path in = Paths.get(resource.toURI());
        Path out = Paths.get("build/output3.txt");
        try {
            AsyncFiles.readAllBytes(in)
                    .thenCompose(bytes -> AsyncFiles.writeBytes(out, bytes))
                    .join();
            byte[] expected = Files.readAllBytes(in);
            byte[] actual = Files.readAllBytes(out);
            assertArrayEquals(expected, actual);
        } finally {
            delete(out);
        }
    }

    @Test
    public void readWriteTestSyncForReadme() throws IOException, URISyntaxException {
        URL resource = getClass().getResource("/Metamorphosis-by-Franz-Kafka.txt");
        Path in = Paths.get(resource.toURI());
        Path out = Paths.get("build/output4.txt");
        try {
            byte[] data = Files.readAllBytes(in);
            Files.write(out, data);
            byte[] expected = Files.readAllBytes(in);
            byte[] actual = Files.readAllBytes(out);
            assertArrayEquals(expected, actual);
        } finally {
            delete(out);
        }
    }
}