package org.xbib.event.async.io;

import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionException;
import java.util.concurrent.ExecutionException;
import java.util.stream.Stream;

import static java.nio.channels.AsynchronousFileChannel.open;
import static java.nio.file.Files.delete;
import static java.util.Arrays.asList;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

public class AsyncFilesFailures {

    private static final Path METAMORPHOSIS = getPath("Metamorphosis-by-Franz-Kafka.txt");

    private static final Path UTF_8_INVALID = getPath("UTF-8-test.txt");

    private static Path getPath(String resource) {
        try {
            return Paths.get(AsyncFileReaderTest.class.getResource("/" + resource).toURI());
        } catch (NullPointerException | URISyntaxException e) {
            e.printStackTrace();
            throw new IllegalStateException();
        }
    }

    private static void openLock(Path output) {
        try {
            open(output, StandardOpenOption.WRITE).lock().get();
        } catch (InterruptedException | IOException | ExecutionException e) {
            throw new RuntimeException(e);
        }
    }

    @Test
    public void readAllNoSuchFile() {
        AsyncFiles.readAllBytes(Paths.get("non-existent-file.sgf"))
                .whenComplete((pos, err) -> {
                    if (err == null)
                        fail("It should fail reading an nonexistent!");
                    else
                        assertEquals(err.getClass(), NoSuchFileException.class);
                });
    }

    @Test
    public void readAllLinesNoSuchFile() {
        AsyncFiles.lines("non-existent-file.sgf")
                .subscribe(Subscribers
                        .doOnNext(line ->
                                fail("It should fail reading an nonexistent!")
                        )
                        .doOnError(err ->
                                assertEquals(err.getClass(), NoSuchFileException.class)
                        ));
    }

    @Test
    public void readAllToCallbackNoSuchFile() {
        AsyncFiles.readAll("non-existent-file.sgf", (err, data) -> {
                    if (err == null || data != null) {
                        fail("It should fail reading an nonexistent!");
                    } else {
                        assertEquals(err.getClass(), NoSuchFileException.class);
                    }
                });
    }

    @Disabled
    @Test
    public void concurrentReadLines() throws IOException {
        final Path path = Paths.get("build/dummy1.txt");
        List<String> expected = asList("super", "brave", "isel", "ole", "gain", "massi", "tot");
        Files.write(path, expected, StandardOpenOption.CREATE);
        try {
            CompletableFuture<Void> promise = new CompletableFuture<>();
            boolean[] done = {false};
            AsyncFiles.lines(8, path)
                    .subscribe(Subscribers
                            .doOnNext(item -> {
                                if (!done[0]) {
                                    openLock(path);
                                    done[0] = true;
                                }
                            })
                            .doOnSubscribe(sign -> sign.request(Integer.MAX_VALUE))
                            .doOnError(err -> {
                                assertTrue(err instanceof IOException);
                                promise.complete(null);
                            }));
            promise.join();
        } finally {
            Files.delete(path);
        }
    }

    @Test
    public void concurrentReadBytes() {
        Path path = METAMORPHOSIS;
        CompletableFuture<byte[]> p = AsyncFiles
                .readAllBytes(path, 8)
                .whenComplete((arr, err) -> {
                    assertNull(arr);
                    assertTrue(err instanceof CompletionException);
                });
        openLock(path);
        try {
            p.join();
            fail("It should be completed exceptionally!");
        } catch (CompletionException e) {
            //
        }
    }

    @Disabled
    @Test
    public void concurrentWriteLines() throws IOException {
        final Path path = Paths.get("build/dummy2.txt");
        Stream<String> lines = Files.lines(METAMORPHOSIS);
        Iterable<String> expected = lines::iterator;
        try {
            CompletableFuture<Integer> p = AsyncFiles.write(path, expected);
            openLock(path);
            p.join();
            fail("It should be completed exceptionally!");
        } catch (CompletionException e) {
            /* Assert true */
        } finally {
            delete(path);
        }
    }

    @Test
    public void concurrentBytesLinesOnOpen() throws IOException {
        final Path path = Paths.get("build/dummy3.txt");
        Files.write(path, "ola".getBytes(), StandardOpenOption.CREATE);
        try {
            AsyncFiles.writeBytes(path, null)
                    .whenComplete((pos, err) -> {
                        if (err == null) {
                            fail("It should fail creating a file that already exists");
                        } else {
                            assertEquals(err.getClass(), IOException.class);
                        }
                    });
        } finally {
            delete(path);
        }
    }

    @Test
    public void concurrentWriteLinesOnOpen() throws IOException {
        final Path path = Paths.get("build/dummy4.txt");
        Files.write(path, "ola".getBytes(), StandardOpenOption.CREATE);
        try {
            AsyncFiles.write(path, null)
                    .whenComplete((pos, err) -> {
                        if (err == null)
                            fail("It should fail creating a file that already exists");
                        else
                            assertEquals(err.getClass(), IOException.class);
                    });
        } finally {
            delete(path);
        }
    }

    @Test
    public void readLinesWithInvalidBytes() {
        CompletableFuture<Void> p = new CompletableFuture<>();
        AsyncFiles
                .lines(4, UTF_8_INVALID)
                .subscribe(Subscribers
                        .doOnNext((item) -> {
                        })
                        .doOnSubscribe(sign -> sign.request(Integer.MAX_VALUE))
                        .doOnError(err -> fail())
                        .doOnComplete(() ->
                                p.complete(null)
                        )
                );
        p.join();
    }
}
