package org.xbib.event.clock;

import org.xbib.event.bus.EventBus;
import org.xbib.event.common.EventManagerService;
import org.xbib.settings.Settings;
import org.xbib.time.schedule.CronExpression;
import org.xbib.time.schedule.CronSchedule;

import java.io.Closeable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadFactory;
import java.util.logging.Level;
import java.util.logging.Logger;

public class ClockEventManagerService implements EventManagerService, Closeable {

    private static final Logger logger = Logger.getLogger(ClockEventManagerService.class.getName());

    private CronSchedule<Integer> cronSchedule;

    private List<String> suspended;

    public ClockEventManagerService() {
    }

    public ClockEventManagerService init(Settings settings, EventBus eventBus, ClassLoader classLoader, ExecutorService executorService) {
        this.suspended = new ArrayList<>();
        ThreadFactory threadFactory = new ThreadFactory() {
            int n = 1;
            @Override
            public Thread newThread(Runnable r) {
                return new Thread(r, "clock-event-manager-" + (n++));
            }
        };
        ScheduledExecutorService scheduledExecutorService =
                Executors.newScheduledThreadPool(settings.getAsInt("pool.size", 2), threadFactory);
        this.cronSchedule = new CronSchedule<>(scheduledExecutorService);
        for (Map.Entry<String, Settings> clockEventService : settings.getGroups("event.clock").entrySet()) {
            String name = clockEventService.getKey();
            Settings entrySettings = clockEventService.getValue();
            if (entrySettings.getAsBoolean("enabled", true)) {
                String entry = entrySettings.get("entry");
                if (entry != null) {
                    try {
                        String className = entrySettings.get("class");
                        if (className != null) {
                            @SuppressWarnings("unchecked")
                            Class<? extends ClockEvent> eventClass = (Class<? extends ClockEvent>) classLoader.loadClass(className);
                            cronSchedule.add(className, CronExpression.parse(entry), new ClockEventService(this, eventBus, name, eventClass));
                            logger.log(Level.INFO, "cron job " + clockEventService.getKey() + " scheduled on " + entry + ", event class " + className);
                        } else {
                            logger.log(Level.WARNING, "no class specified");
                        }
                    } catch (Exception e) {
                        logger.log(Level.WARNING, "unable to schedule cron job " + clockEventService.getKey() + ", reason " + e.getMessage());
                    }
                }
            } else {
                logger.log(Level.WARNING, "clock event service " + name + " in configuration not enabled");
            }
        }
        logger.log(Level.INFO, "entries = " + cronSchedule.getEntries());
        cronSchedule.start();
        return this;
    }

    public List<String> getSuspended() {
        return suspended;
    }

    public void suspend(String name) {
        suspended.add(name);
    }

    public void resume(String name) {
        suspended.remove(name);
    }

    @Override
    public void close() throws IOException {
        cronSchedule.close();
    }
}
