package org.xbib.event.common;

import java.util.HashMap;
import java.util.Map;
import java.util.ServiceLoader;

import org.xbib.event.Event;
import org.xbib.event.EventConsumer;
import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.SubscriberExceptionContext;
import org.xbib.event.bus.SubscriberExceptionHandler;
import org.xbib.event.clock.ClockEventManagerService;
import org.xbib.event.generic.GenericEventManagerService;
import org.xbib.event.path.FileFollowEventManagerService;
import org.xbib.event.path.PathEventManagerService;
import org.xbib.event.timer.TimerEventManagerService;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class EventManager {

    private static final Logger logger = Logger.getLogger(EventManager.class.getName());

    private final Builder builder;

    private final Map<Class<? extends EventManagerService>, EventManagerService> services;

    private EventManager(Builder builder) {
        this.builder = builder;
        this.services = new HashMap<>();
        services.put(GenericEventManagerService.class, new GenericEventManagerService()
                .init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        services.put(ClockEventManagerService.class, new ClockEventManagerService()
                .init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        services.put(TimerEventManagerService.class, new TimerEventManagerService()
                .init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        services.put(FileFollowEventManagerService.class, new FileFollowEventManagerService()
                .init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        services.put(PathEventManagerService.class, new PathEventManagerService()
                .init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        for (EventManagerService service : ServiceLoader.load(EventManagerService.class)) {
            services.put(service.getClass(), service.init(builder.settings, builder.eventBus, builder.classLoader, builder.executorService));
        }
        logger.log(Level.INFO, "installed event service managers = " + services.keySet());
    }

    public static Builder builder(Settings settings) {
        return new Builder(settings);
    }

    public void submit(Event event) {
        getGenericEventManagerService().post(event);
    }

    public EventManagerService getEventManagerService(Class<? extends EventManagerService> cl) {
        return services.get(cl);
    }

    public GenericEventManagerService getGenericEventManagerService() {
        return (GenericEventManagerService) services.get(GenericEventManagerService.class);
    }

    public ClockEventManagerService getClockEventManagerService() {
        return (ClockEventManagerService) services.get(ClockEventManagerService.class);
    }

    public TimerEventManagerService getTimerEventManagerService() {
        return (TimerEventManagerService) services.get(TimerEventManagerService.class);
    }

    public FileFollowEventManagerService getFileFollowEventManagerService() {
        return (FileFollowEventManagerService) services.get(FileFollowEventManagerService.class);
    }

    public PathEventManagerService getPathEventManagerService() {
        return (PathEventManagerService) services.get(PathEventManagerService.class);
    }

    public void close() throws IOException {
        for (EventConsumer eventConsumer : builder.eventConsumers) {
            if (eventConsumer instanceof Closeable closeable) {
                closeable.close();
            }
        }
        for (EventManagerService service : services.values()) {
            if (service instanceof Closeable closeable) {
                closeable.close();
            }
        }
    }

    public static class Builder {

        private final Settings settings;

        private ExecutorService executorService;

        private ClassLoader classLoader;

        private SubscriberExceptionHandler subscriberExceptionHandler;

        private AsyncEventBus eventBus;

        private final List<EventConsumer> eventConsumers;

        private Builder(Settings settings) {
            this.settings = settings;
            this.eventConsumers = new ArrayList<>();
            this.subscriberExceptionHandler = new EventManagerExceptionHandler();
        }

        public Builder setExecutorService(ExecutorService executorService) {
            this.executorService = executorService;
            return this;
        }

        public Builder setClassLoader(ClassLoader classLoader) {
            this.classLoader = classLoader;
            return this;
        }

        public Builder setEventBus(AsyncEventBus eventBus) {
            this.eventBus = eventBus;
            return this;
        }

        public Builder register(EventConsumer eventConsumer) {
            Objects.requireNonNull(eventConsumer, "event consumer must not be null");
            eventConsumers.add(eventConsumer);
            return this;
        }

        public Builder setSubscriberExceptionHandler(SubscriberExceptionHandler subscriberExceptionHandler) {
            this.subscriberExceptionHandler = subscriberExceptionHandler;
            return this;
        }

        public Builder loadEventConsumers() {
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            for (EventConsumer eventConsumer : ServiceLoader.load(EventConsumer.class, classLoader)) {
                eventConsumers.add(eventConsumer);
            }
            return this;
        }

        public EventManager build() {
            if (executorService == null) {
                executorService = Executors.newFixedThreadPool(settings.getAsInt("event.threadcount", 2));
            }
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            if (eventBus == null) {
                eventBus = new AsyncEventBus(executorService, subscriberExceptionHandler);
            }
            for (EventConsumer eventConsumer : eventConsumers) {
                eventBus.register(eventConsumer);
            }
            return new EventManager(this);
        }
    }

    private static class EventManagerExceptionHandler implements SubscriberExceptionHandler {
        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            logger.log(Level.SEVERE, exception.getMessage(), exception);
        }
    }
}
