package org.xbib.event.path;

import org.xbib.datastructures.api.TimeValue;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.bus.EventBus;
import org.xbib.event.common.EventManagerService;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class PathEventManagerService implements EventManagerService, Closeable {

    private static final Logger logger = Logger.getLogger(PathEventManagerService.class.getName());

    private EventBus eventBus;

    private ExecutorService executorService;

    private Path path;

    private Map<Future<?>, PathEventService> eventServiceMap;

    private List<String> suspendedQueues;

    public PathEventManagerService() {
    }

    @SuppressWarnings("unchecked")
    @Override
    public PathEventManagerService init(Settings settings,
                                        EventBus eventBus,
                                        ClassLoader classLoader,
                                        ExecutorService executorService) {
        this.eventBus = eventBus;
        this.executorService = executorService;
        this.eventServiceMap = new LinkedHashMap<>();
        this.suspendedQueues = new ArrayList<>();
        this.path = Paths.get(settings.get("event.pathevent.base", "/var/tmp/pathevent"));
        for (Map.Entry<String, Settings> entry : settings.getGroups("event.path").entrySet()) {
            try {
                String name = entry.getKey();
                Settings definition = entry.getValue();
                if (definition.getAsBoolean("enabled", true)) {
                    int maxBytes = definition.getAsInt("maxfilesize", 10 * 1024 * 1024); // 10 MB
                    TimeValue lifetime = definition.getAsTime("lifetime", TimeValue.timeValueHours(72));
                    String className = definition.get("class", PathEventImpl.class.getName());
                    Class<? extends PathEvent> eventClass = (Class<? extends PathEvent>) classLoader.loadClass(className);
                    Path p = path.resolve(name);
                    createPathEventService(name, p, maxBytes, lifetime, eventClass);
                } else {
                    logger.log(Level.WARNING, "path servive definition not enabled in configuration");
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, "unable to create event path service " + entry.getKey() + ", reason " + e.getMessage(), e);
            }
        }
        return this;
    }

    public void createPathEventService(String name,
                                       Path path,
                                       int maxBytes,
                                       TimeValue lifetime,
                                       Class<? extends PathEvent> eventClass)
            throws IOException {
        createQueue(name, path);
        PathEventService pathEventService = new PathEventService(this, eventBus, name, path, maxBytes, lifetime, eventClass);
        add(pathEventService);
    }

    public void add(PathEventService pathEventService) {
        Future<?> future = executorService.submit(pathEventService);
        eventServiceMap.put(future, pathEventService);
        logger.log(Level.INFO, "path event service " + pathEventService + " added");
    }

    @Override
    public void close() throws IOException {
        logger.log(Level.INFO, "closing all path event services");
        eventServiceMap.forEach((k, v) -> {
            k.cancel(true);
            try {
                v.close();
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        });
    }

    public List<String> getSuspendedQueues() {
        return suspendedQueues;
    }

    public void suspend(String queue) {
        suspendedQueues.add(queue);
    }

    public void resume(String queue) {
        suspendedQueues.remove(queue);
    }

    public boolean put(String queue, String key, Map<String,Object> map) throws IOException {
         return put(queue, key, ".json", Json.toString(map));
    }

    public boolean putIfNotExists(String queue, String key, Map<String,Object> map) throws IOException {
        if (!exists(queue, key, ".json")) {
            return put(queue, key, ".json", Json.toString(map));
        } else {
            return false;
        }
    }

    public boolean put(String queue, String key, String suffix, String string) throws IOException {
        String keyFileName = key + suffix;
        Path queuePath = path.resolve(queue);
        if (Files.exists(queuePath.resolve(PathEventService.INCOMING).resolve(keyFileName)) ||
                Files.exists(queuePath.resolve(PathEventService.SUCCESS).resolve(keyFileName))) {
            logger.log(Level.WARNING, "key " + key + " already exists");
            return false;
        }
        Path p = queuePath.resolve(PathEventService.INCOMING).resolve(keyFileName);
        try (Writer writer = Files.newBufferedWriter(p)) {
            writer.write(string);
        }
        // obligatory purge. This is hacky.
        eventServiceMap.forEach((k, v) -> {
            if (v.getName().equals(queue)) {
                try {
                    v.purge();
                } catch (IOException e) {
                    throw new UncheckedIOException(e);
                }
            }
        });
        return true;
    }

    public boolean exists(String queue, String key, String suffix) {
        String keyFileName = key + suffix;
        Path queuePath = path.resolve(queue);
        return Files.exists(queuePath.resolve(PathEventService.INCOMING).resolve(keyFileName)) ||
                Files.exists(queuePath.resolve(PathEventService.SUCCESS).resolve(keyFileName));
    }

    public long sizeOfIncoming(String queue) throws IOException {
        return sizeOf(path.resolve(queue).resolve(PathEventService.INCOMING));
    }

    public long sizeOfSuccess(String queue) throws IOException {
        return sizeOf(path.resolve(queue).resolve(PathEventService.SUCCESS));
    }

    public long sizeOfFail(String queue) throws IOException {
        return sizeOf(path.resolve(queue).resolve(PathEventService.FAIL));
    }

    public static long sizeOf(Path path) throws IOException {
        try (Stream<Path> stream = Files.find(path, 1, (p, basicFileAttributes) -> Files.isRegularFile(p))) {
            return stream.count();
        }
    }

    private static void createQueue(String name, Path p) throws IOException {
        if (!Files.exists(p)) {
            logger.log(Level.FINE, "creating queue " + name + " at " + p);
            Files.createDirectories(p);
            for (String s: List.of(PathEventService.INCOMING, PathEventService.SUCCESS, PathEventService.FAIL)) {
                Path dir = p.resolve(s);
                if (!Files.exists(dir)) {
                    logger.log(Level.FINE, "creating queue " + name + " dir " + dir);
                    Files.createDirectories(dir);
                }
            }
        }
    }
}
