package org.xbib.event.path;

import org.xbib.datastructures.api.TimeValue;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.bus.EventBus;

import java.io.Closeable;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.ClosedWatchServiceException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PathEventService implements Callable<Integer>, Closeable {

    private static final Logger logger = Logger.getLogger(PathEventService.class.getName());

    public static final String INCOMING = "incoming";

    public static final String SUCCESS = "success";

    public static final String FAIL = "fail";

    private final PathEventManagerService pathEventManager;

    private final EventBus eventBus;

    private final Path path;

    private final String name;

    private final int maxFileSize;

    private final TimeValue lifetime;
;
    private final Class<? extends PathEvent> pathEventClass;

    private final WatchService watchService;

    private int eventCount;

    private volatile boolean keepWatching;

    public PathEventService(PathEventManagerService pathEventManager,
                            EventBus eventBus,
                            String name,
                            Path path,
                            int maxFileSize,
                            TimeValue lifetime,
                            Class<? extends PathEvent> pathEventClass) throws IOException {
        this.pathEventManager = pathEventManager;
        this.eventBus = eventBus;
        this.name = name;
        this.path = path;
        this.maxFileSize = maxFileSize;
        this.lifetime = lifetime;
        this.pathEventClass = pathEventClass;
        drainIncoming();
        this.watchService = path.getFileSystem().newWatchService();
        WatchEvent.Kind<?>[] kinds = new WatchEvent.Kind[] { StandardWatchEventKinds.ENTRY_CREATE };
        WatchKey watchKey = path.resolve(INCOMING).register(watchService, kinds);
        keepWatching = true;
        logger.log(Level.INFO, "path event service created for incoming files at " + path + " max file size = " + maxFileSize);
    }

    public String getName() {
        return name;
    }


    @SuppressWarnings("unchecked")
    @Override
    public Integer call() {
        try {
            logger.log(Level.INFO, "watch service running on " + path.resolve(INCOMING));
            while (keepWatching && watchService != null) {
                WatchKey watchKey = watchService.take();
                logger.log(Level.FINE, "received a watch key " + watchKey);
                for (WatchEvent<?> watchEvent : watchKey.pollEvents()) {
                    WatchEvent.Kind<?> kind = watchEvent.kind();
                    if (kind == StandardWatchEventKinds.OVERFLOW) {
                        continue;
                    }
                    // we sleep here a bit, to give time to the OS to complete file writing
                    Thread.sleep(1000L);
                    WatchEvent<Path> pathWatchEvent = (WatchEvent<Path>) watchEvent;
                    String watchEventContext = pathWatchEvent.context().toString();
                    Path p = path.resolve(INCOMING).resolve(watchEventContext);
                    logger.log(Level.FINE, "watch event " + pathWatchEvent + " context = " + watchEventContext + " path = " + p);
                    if (pathEventManager.getSuspendedQueues().contains(name)) {
                        failEvent(watchEventContext, p);
                    } else {
                        postEvent(watchEventContext, p);
                    }
                }
                watchKey.reset();
            }
        } catch (ClosedWatchServiceException e) {
            logger.log(Level.SEVERE, "closed watch key: " + e.getMessage(), e);
        } catch (InterruptedException e) {
            logger.log(Level.WARNING, "interrupted");
            Thread.currentThread().interrupt();
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return eventCount;
    }

    @Override
    public void close() throws IOException {
        keepWatching = false;
        if (watchService != null) {
            logger.log(Level.FINE, "closing watch service " + watchService);
            watchService.close();
        }
    }

    public void drainIncoming() throws IOException {
        try (DirectoryStream<Path> stream = Files.newDirectoryStream(path.resolve(INCOMING))) {
            stream.forEach(path -> {
                if (Files.isRegularFile(path)) {
                    String key = path.getFileName().toString();
                    logger.log(Level.INFO, "while draining found key = " + key + " path = " + path);
                    postEvent(key, path);
                }
            });
        }
    }

    public void purge() throws IOException {
        purge(path);
    }

    public void purge(Path path) throws IOException {
        try (DirectoryStream<Path> stream = Files.newDirectoryStream(path)) {
            stream.forEach(p -> {
                try {
                    if (Files.isRegularFile(p)) {
                        BasicFileAttributeView view = Files.getFileAttributeView(p, BasicFileAttributeView.class);
                        BasicFileAttributes attrs = view.readAttributes();
                        if (Instant.now().minusMillis(attrs.lastModifiedTime().toMillis()).toEpochMilli() > lifetime.millis()) {
                            logger.log(Level.WARNING, "lifetime of " + lifetime + " exceeded, deleting " + p);
                            Files.delete(p);
                        }
                    } else {
                        purge(p);
                    }
                } catch (IOException e) {
                    throw new UncheckedIOException(e);
                }
            });
        }
    }

    private void postEvent(String key, Path file) {
        PathEvent event = toEvent(key, file);
        if (event != null) {
            logger.log(Level.FINE, "posting new event = " + event.getClass());
            eventBus.post(event);
            eventCount++;
        }
    }

    private void failEvent(String key, Path file) throws IOException {
        PathEvent event = toEvent(key, file);
        if (event != null) {
            logger.log(Level.WARNING, "queue " + name + " suspended, event short-circuited to fail");
            event.fail();
        }
    }

    private static String getBase(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(0, pos) : name;
    }

    private static String getSuffix(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(pos + 1) : null;
    }

    private PathEvent toEvent(String key, Path file) {
        try {
            String base = getBase(key);
            String suffix = getSuffix(key);
            long fileSize = Files.size(file);
            if (fileSize > maxFileSize) {
                logger.log(Level.SEVERE, "event object ignored, too large");
                return null;
            }
            if ("!json".equals(suffix)) {
                logger.log(Level.SEVERE, "event object ignored, no json suffix");
                return null;
            }
            String json = Files.readString(file);
            return toEvent(base, file, suffix, json);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        logger.log(Level.SEVERE, "event object could not be created");
        return null;
    }

    private PathEvent toEvent(String base, Path file, String suffix, String json) {
        try {
            PathEvent event = pathEventClass.getConstructor().newInstance();
            event.setType(base);
            event.setFile(file);
            event.setSuffix(suffix);
            event.setPath(path); // remember directory for fail() and success()
            event.setMap(Json.toMap(json));
            return event;
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return null;
        }
    }
}
