package org.xbib.event.timer;

import org.xbib.event.bus.EventBus;
import org.xbib.event.persistence.PersistenceStore;

import java.io.Closeable;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

class TimerEventService implements Closeable {

    private static final Logger logger = Logger.getLogger(TimerEventService.class.getName());
    
    private final EventBus eventBus;

    private final Class<? extends TimerEvent> eventClass;

    private final ZoneId zoneId;

    private final PersistenceStore<String, Object> persistenceStore;

    private final Timer timer;

    public TimerEventService(EventBus eventBus,
                             String name,
                             Class<? extends TimerEvent> eventClass,
                             ZoneId zoneId,
                             PersistenceStore<String, Object> persistenceStore) throws IOException {
        this.eventBus = eventBus;
        this.eventClass = eventClass;
        this.zoneId = zoneId;
        this.persistenceStore = persistenceStore;
        this.timer = new Timer();
        int tasknum = reschedule();
        logger.log(Level.INFO, "timer event service " + name + " loaded and rescheduled, " +  tasknum + " timer tasks");
    }

    void schedule(Instant instant, Map<String, Object> map) throws IOException {
        ZonedDateTime zonedDateTime = instant.atZone(zoneId);
        Map<String, Object> task = new LinkedHashMap<>(map);
        task.put("scheduled", zonedDateTime.format(DateTimeFormatter.ISO_DATE_TIME));
        TimerEventTask timerEventTask = new TimerEventTask(task);
        timer.schedule(timerEventTask, Date.from(instant));
        logger.log(Level.INFO, "new task " + map + " added, scheduled at " + instant);
    }

    @SuppressWarnings("unchecked")
    int reschedule() throws IOException {
        logger.log(Level.INFO, "rescheduling timer events");
        persistenceStore.load();
        List<Map<String, Object>> tasks = (List<Map<String, Object>>) persistenceStore.getOrDefault("tasks", new ArrayList<>());
        persistenceStore.clear();
        persistenceStore.commit();
        for (Map<String, Object> task : tasks) {
            ZonedDateTime scheduledDate = ZonedDateTime.parse((String) task.get("scheduled"), DateTimeFormatter.ISO_DATE_TIME);
            if (scheduledDate.isBefore(ZonedDateTime.now())) {
                logger.log(Level.WARNING, "scheduled timer task " + task + " date already passed");
            } else {
                schedule(scheduledDate.toInstant(), task);
            }
        }
        tasks = (List<Map<String, Object>>) persistenceStore.getOrDefault("tasks", new ArrayList<>());
        logger.log(Level.INFO, "rescheduling timer events complete: " + tasks.size() + " events");
        return tasks.size();
    }

    @Override
    public void close() throws IOException {
        timer.cancel();
        persistenceStore.commit();
    }

    void purge() {
        timer.purge();
    }

    public class TimerEventTask extends TimerTask {

        private final Map<String, Object> map;

        public TimerEventTask(Map<String, Object> map) throws IOException {
            this.map = map;
            persistenceStore.insert("tasks", this.map);
        }

        @Override
        public void run() {
            TimerEvent timerEvent;
            try {
                timerEvent = eventClass.getDeclaredConstructor().newInstance();
                timerEvent.setInstant(Instant.now());
                timerEvent.setMap(map);
                logger.log(Level.FINE, "posting timer event " + timerEvent.getClass().getName() + " map = " + map);
                eventBus.post(timerEvent);
                logger.log(Level.FINE, "persistence before remove: " + persistenceStore.get("tasks"));
                if (persistenceStore.remove("tasks", this.map)) {
                    logger.log(Level.FINE, "removal done");
                }
                logger.log(Level.FINE, "persistence after remove: " + persistenceStore.get("tasks"));
            } catch (InstantiationException | IllegalAccessException | InvocationTargetException | NoSuchMethodException | IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        @Override
        public String toString() {
            return map.toString();
        }

        @Override
        public int hashCode() {
            return map.hashCode();
        }

        @Override
        public boolean equals(Object object) {
            return object instanceof TimerEventTask && Objects.equals(map, ((TimerEventTask) object).map);
        }
    }
}
