package org.xbib.event.bus;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.fail;

public class EventBusTest {

    private static final Logger logger = Logger.getLogger(EventBusTest.class.getName());

    private static final String EVENT = "Hello";

    private static final String BUS_IDENTIFIER = "test-bus";

    private EventBus bus;

    @BeforeEach
    public void create() {
        bus = new EventBus(BUS_IDENTIFIER);
    }

    @Test
    public void testPolymorphicDistribution() {
        // Three catchers for related types String, Object, and Comparable<?>.
        // String isa Object
        // String isa Comparable<?>
        // Comparable<?> isa Object
        StringCatcher stringCatcher = new StringCatcher();

        final List<Object> objectEvents = new ArrayList<>();
        Object objCatcher =
                new Object() {
                    @SuppressWarnings("unused")
                    @Subscribe
                    public void eat(Object food) {
                        objectEvents.add(food);
                    }
                };

        final List<Comparable<?>> compEvents = new ArrayList<>();
        Object compCatcher =
                new Object() {
                    @SuppressWarnings("unused")
                    @Subscribe
                    public void eat(Comparable<?> food) {
                        compEvents.add(food);
                    }
                };
        bus.register(stringCatcher);
        bus.register(objCatcher);
        bus.register(compCatcher);

        // Two additional event types: Object and Comparable<?> (played by Integer)
        Object objEvent = new Object();
        Object compEvent = 6;

        bus.post(EVENT);
        bus.post(objEvent);
        bus.post(compEvent);

        List<String> stringEvents = stringCatcher.getEvents();

        logger.log(Level.INFO, stringEvents.toString());

        assertEquals(1, stringEvents.size(), "Only one String should be delivered.");
        assertEquals(EVENT, stringEvents.get(0), "Correct string should be delivered.");

        // Check the Catcher<Object>...
        assertEquals(3, objectEvents.size(), "Three Objects should be delivered.");
        assertEquals(EVENT, objectEvents.get(0), "String fixture must be first object delivered.");
        assertEquals(objEvent, objectEvents.get(1), "Object fixture must be second object delivered.");
        assertEquals(compEvent, objectEvents.get(2), "Comparable fixture must be thirdobject delivered.");

        // Check the Catcher<Comparable<?>>...
        assertEquals(2, compEvents.size(), "Two Comparable<?>s should be delivered.");
        assertEquals(EVENT, compEvents.get(0), "String fixture must be first comparable delivered.");
        assertEquals(compEvent, compEvents.get(1), "Comparable fixture must be second comparable delivered.");
    }

    @Test
    public void testSubscriberThrowsException() throws Exception {
        final RecordingSubscriberExceptionHandler handler = new RecordingSubscriberExceptionHandler();
        final EventBus eventBus = new EventBus(handler);
        final RuntimeException exception =
                new RuntimeException("but culottes have a tendancy to ride up!");
        final Object subscriber =
                new Object() {
                    @Subscribe
                    public void throwExceptionOn(String message) {
                        throw exception;
                    }
                };
        eventBus.register(subscriber);
        eventBus.post(EVENT);

        assertNotNull(handler.getContext());
        assertEquals(exception, handler.getException(), "Cause should be available.");
        assertEquals(eventBus, handler.getContext().getEventBus(), "EventBus should be available.");
        assertEquals(EVENT, handler.getContext().getEvent(), "Event should be available.");
        assertEquals(subscriber, handler.getContext().getSubscriber(), "Subscriber should be available.");
        assertEquals(subscriber.getClass().getMethod("throwExceptionOn", String.class),
                handler.getContext().getSubscriberMethod(),"Method should be available.");
    }

    static class RecordingSubscriberExceptionHandler implements SubscriberExceptionHandler {

        private SubscriberExceptionContext context;

        private Throwable exception;

        public RecordingSubscriberExceptionHandler() {
        }

        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            this.exception = exception;
            this.context = context;
        }

        public SubscriberExceptionContext getContext() {
            return context;
        }

        public Throwable getException() {
            return exception;
        }
    }

    @Test
    public void testSubscriberThrowsExceptionHandlerThrowsException() {
        final EventBus eventBus = new EventBus((exception, context) -> {
                            throw new RuntimeException();
                        });
        final Object subscriber =
                new Object() {
                    @Subscribe
                    public void throwExceptionOn(String message) {
                        throw new RuntimeException();
                    }
                };
        eventBus.register(subscriber);
        try {
            eventBus.post(EVENT);
        } catch (RuntimeException e) {
            fail("Exception should not be thrown.");
        }
    }

    @Test
    public void testDeadEventForwarding() {
        GhostCatcher catcher = new GhostCatcher();
        bus.register(catcher);
        bus.post(EVENT);
        List<DeadEvent> events = catcher.getEvents();
        assertEquals(1, events.size(), "One dead event should be delivered.");
        assertEquals(EVENT, events.get(0).getEvent(), "The dead event should wrap the original event.");
    }

    static class GhostCatcher {

        private final List<DeadEvent> events = new ArrayList<>();

        @Subscribe
        public void ohNoesIHaveDied(DeadEvent event) {
            events.add(event);
        }

        public List<DeadEvent> getEvents() {
            return events;
        }
    }


    @Test
    public void testMissingSubscribe() {
        bus.register(new Object());
    }

    @Test
    public void testUnregister() {
        StringCatcher catcher1 = new StringCatcher();
        StringCatcher catcher2 = new StringCatcher();
        try {
            bus.unregister(catcher1);
            fail("Attempting to unregister an unregistered object succeeded");
        } catch (IllegalArgumentException expected) {
            // OK.
        }
        bus.register(catcher1);
        bus.post(EVENT);
        bus.register(catcher2);
        bus.post(EVENT);
        List<String> expectedEvents = new ArrayList<>();
        expectedEvents.add(EVENT);
        expectedEvents.add(EVENT);
        assertEquals(expectedEvents, catcher1.getEvents(), "Two correct events should be delivered.");
        assertEquals(List.of(EVENT), catcher2.getEvents(), "One correct event should be delivered.");
        bus.unregister(catcher1);
        bus.post(EVENT);
        assertEquals(expectedEvents, catcher1.getEvents(), "Shouldn't catch any more events when unregistered.");
        assertEquals(expectedEvents, catcher2.getEvents(), "Two correct events should be delivered.");
        try {
            bus.unregister(catcher1);
            fail("Attempting to unregister an unregistered object succeeded");
        } catch (IllegalArgumentException expected) {
            // OK.
        }
        bus.unregister(catcher2);
        bus.post(EVENT);
        assertEquals(expectedEvents, catcher1.getEvents(), "Shouldn't catch any more events when unregistered.");
        assertEquals(expectedEvents, catcher2.getEvents(),  "Shouldn't catch any more events when unregistered.");
    }

    // NOTE: This test will always pass if register() is thread-safe but may also
    // pass if it isn't, though this is unlikely.
    @Test
    public void testRegisterThreadSafety() throws Exception {
        List<StringCatcher> catchers = new CopyOnWriteArrayList<>();
        List<Future<?>> futures = new ArrayList<>();
        ExecutorService executor = Executors.newFixedThreadPool(10);
        int numberOfCatchers = 10000;
        for (int i = 0; i < numberOfCatchers; i++) {
            futures.add(executor.submit(new Registrator(bus, catchers)));
        }
        for (int i = 0; i < numberOfCatchers; i++) {
            futures.get(i).get();
        }
        assertEquals(numberOfCatchers, catchers.size(), "Unexpected number of catchers in the list");
        bus.post(EVENT);
        List<String> expectedEvents = List.of(EVENT);
        for (StringCatcher catcher : catchers) {
            assertEquals(expectedEvents, catcher.getEvents(), "One of the registered catchers did not receive an event.");
        }
    }

    /** Runnable which registers a StringCatcher on an event bus and adds it to a list. */
    private static class Registrator implements Runnable {
        private final EventBus bus;
        private final List<StringCatcher> catchers;

        Registrator(EventBus bus, List<StringCatcher> catchers) {
            this.bus = bus;
            this.catchers = catchers;
        }

        @Override
        public void run() {
            StringCatcher catcher = new StringCatcher();
            bus.register(catcher);
            catchers.add(catcher);
        }
    }

    @Test
    public void testToString() {
        EventBus eventBus = new EventBus("a b ; - \" < > / \\ €");
        assertEquals("EventBus{a b ; - \" < > / \\ €}", eventBus.toString());
    }

    /**
     * Tests that bridge methods are not subscribed to events. In Java 8, annotations are included on
     * the bridge method in addition to the original method, which causes both the original and bridge
     * methods to be subscribed (since both are annotated @Subscribe) without specifically checking
     * for bridge methods.
     */
    @Test
    public void testRegistrationWithBridgeMethod() {
        final AtomicInteger calls = new AtomicInteger();
        bus.register(
                new Callback<String>() {
                    @Subscribe
                    @Override
                    public void call(String s) {
                        calls.incrementAndGet();
                    }
                });
        bus.post("hello");
        assertEquals(1, calls.get());
    }

    private interface Callback<T> {
        void call(T t);
    }

    @Test
    public void testPrimitiveSubscribeFails() {
        class SubscribesToPrimitive {
            @Subscribe
            public void toInt(int i) {}
        }
        try {
            bus.register(new SubscribesToPrimitive());
            fail("should have thrown");
        } catch (IllegalArgumentException expected) {
        }
    }
}
