package org.xbib.event.syslog;

import org.xbib.datastructures.api.Builder;
import org.xbib.datastructures.json.tiny.JsonBuilder;

import java.io.IOException;
import java.time.DateTimeException;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public abstract class MessageParser {

    private static final Logger log = Logger.getLogger(MessageParser.class.getName());

    private static final String NULL_TOKEN = "-";

    protected final List<DateTimeFormatter> dateFormatters;

    private final ThreadLocal<Matcher> matcherStructuredData;

    private final ThreadLocal<Matcher> matcherKeyValue;

    private final ZoneId zoneId;

    public MessageParser() {
        this(ZoneId.of("UTC"));
    }

    public MessageParser(ZoneId zoneId) {
        this.zoneId = zoneId;
        this.dateFormatters = Arrays.asList(DateTimeFormatter.ISO_OFFSET_DATE_TIME,
                new DateTimeFormatterBuilder()
                        .appendPattern("MMM dd")
                        .optionalStart()
                        .appendPattern("[ yyyy]")
                        .parseDefaulting(ChronoField.YEAR_OF_ERA, 1)
                        .optionalEnd()
                        .appendPattern(" HH:mm:ss")
                        .toFormatter(Locale.ROOT)
        );
        this.matcherStructuredData = initMatcher("\\[([^\\]]+)\\]");
        this.matcherKeyValue = initMatcher("(?<key>\\S+)=\"(?<value>[^\"]+)\"|(?<id>\\S+)");
    }

    /**
     * Method is used to parse an incoming syslog message.
     *
     * @param request Incoming syslog request.
     * @return Object to pass along the pipeline. Null if could not be parsed.
     */
    public abstract Message parse(SyslogRequest request) throws IOException;

    protected final ThreadLocal<Matcher> initMatcher(String pattern) {
        return initMatcher(pattern, 0);
    }

    protected final ThreadLocal<Matcher> initMatcher(String pattern, int flags) {
        final Pattern p = Pattern.compile(pattern, flags);
        return new MatcherInheritableThreadLocal(p);
    }

    protected String nullableString(String groupText) {
        return NULL_TOKEN.equals(groupText) ? null : groupText;
    }

    protected LocalDateTime parseDate(String date) {
        final String cleanDate = date.replaceAll("\\s+", " ");
        LocalDateTime result = null;
        for (DateTimeFormatter formatter : dateFormatters) {
            try {
                TemporalAccessor temporal = formatter.parseBest(cleanDate, OffsetDateTime::from, LocalDateTime::from);
                if (temporal instanceof LocalDateTime) {
                    result = ((LocalDateTime) temporal);
                } else {
                    result = ((OffsetDateTime) temporal).withOffsetSameInstant(ZoneOffset.UTC).toLocalDateTime();
                }
                if (result.getLong(ChronoField.YEAR_OF_ERA) == 1) {
                    result = result.withYear(LocalDateTime.now(this.zoneId).getYear());
                }
                break;
            } catch (DateTimeException e) {
                // ignore
            }
        }
        if (result == null) {
            log.log(Level.SEVERE, "could not parse date " + cleanDate);
        }
        return result;
    }

    protected List<Builder> parseStructuredData(String structuredData) throws IOException {
        final Matcher matcher = matcherStructuredData.get().reset(structuredData);
        final List<Builder> result = new ArrayList<>();
        while (matcher.find()) {
            final String input = matcher.group(1);
            Builder builder = JsonBuilder.builder();
            final Matcher kvpMatcher = matcherKeyValue.get().reset(input);
            while (kvpMatcher.find()) {
                final String key = kvpMatcher.group("key");
                final String value = kvpMatcher.group("value");
                final String id = kvpMatcher.group("id");
                if (null != id && !id.isEmpty()) {
                    builder.field("id", id);
                } else {
                    builder.fieldIfNotNull(key, value);
                }
            }
            result.add(builder);
        }
        return result;
    }

    static class MatcherInheritableThreadLocal extends InheritableThreadLocal<Matcher> {
        private final Pattern pattern;

        MatcherInheritableThreadLocal(Pattern pattern) {
            this.pattern = pattern;
        }

        @Override
        protected Matcher initialValue() {
            return this.pattern.matcher("");
        }
    }

}
