package org.xbib.event.syslog;

import io.netty.bootstrap.Bootstrap;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.DatagramChannel;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioDatagramChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.codec.DelimiterBasedFrameDecoder;
import io.netty.handler.codec.Delimiters;
import org.xbib.event.bus.EventBus;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

public class SyslogService implements Closeable {

    private final int port;

    private final Handler handler;

    private final EventBus eventBus;

    private EventLoopGroup group;

    private EventLoopGroup bossGroup;

    private EventLoopGroup workerGroup;

    private ChannelFuture channelFuture;

    public SyslogService(Settings settings,
                         EventBus eventBus) {
        this.port = settings.getAsInt("port", 1514);
        this.handler = new Handler();
        this.eventBus = eventBus;
    }

    public void startUdp() throws InterruptedException {
        group = new NioEventLoopGroup();
        Bootstrap bootstrap = new Bootstrap();
        bootstrap.group(group)
                .channel(NioDatagramChannel.class)
                .option(ChannelOption.SO_BROADCAST, true)
                .handler(new ChannelInitializer<DatagramChannel>() {
                    @Override
                    protected void initChannel(DatagramChannel datagramChannel) {
                        ChannelPipeline channelPipeline = datagramChannel.pipeline();
                        channelPipeline.addLast(
                                new UDPSyslogMessageDecoder(),
                                new SyslogMessageHandler(),
                                handler
                        );
                    }
                });
        this.channelFuture = bootstrap.bind(port).sync();
    }

    public void startTcp() throws InterruptedException {
        bossGroup = new NioEventLoopGroup();
        workerGroup = new NioEventLoopGroup();
        ServerBootstrap serverBootstrap= new ServerBootstrap();
        serverBootstrap.group(bossGroup, workerGroup)
                .channel(NioServerSocketChannel.class)
                .childHandler(new ChannelInitializer<SocketChannel>() {
                    @Override
                    public void initChannel(SocketChannel ch) {
                        ch.pipeline().addLast(
                                new DelimiterBasedFrameDecoder(2000, true, Delimiters.lineDelimiter()),
                                new TCPSyslogMessageDecoder(),
                                new SyslogMessageHandler(),
                                handler
                        );
                    }
                })
                .option(ChannelOption.SO_BACKLOG, 128)
                .childOption(ChannelOption.SO_KEEPALIVE, true);
        this.channelFuture = serverBootstrap.bind(port).sync();
    }

    @Override
    public void close() throws IOException {
        if (group != null) {
            group.shutdownGracefully();
        }
        if (bossGroup != null) {
            bossGroup.shutdownGracefully();
        }
        if (workerGroup != null) {
            workerGroup.shutdownGracefully();
        }
        if (channelFuture != null) {
            channelFuture.channel().close();
        }
    }

    @ChannelHandler.Sharable
    private class Handler extends SimpleChannelInboundHandler<Message> {

        private static final Logger logger = Logger.getLogger(Handler.class.getName());

        public Handler() {
        }

        @Override
        protected void channelRead0(ChannelHandlerContext ctx, Message msg) {
            if (eventBus != null) {
                eventBus.post(msg);
            } else {
                logger.log(Level.INFO, msg.toString());
            }
        }
    }
}
